from enum import Enum, auto
from typing import Any, Dict, List, Union

from flet_core import Column, Container, Row, Stack, alignment

Element = Dict[str, Any]
Options = Union[List[str], str]
Result = Dict[str, Any]


class Size(Enum):
    """Element size options"""

    Small = "small"
    Medium = "medium"
    Large = "large"


class Icon(Enum):
    """Icon variants"""

    Success = "success"
    Warning = "warning"
    Failure = "failure"


class WindowLocation(Enum):
    """A relative location for placing elements or windows"""

    Center = auto()
    TopLeft = auto()


class VerticalLocation(Enum):
    """A vertical location"""

    Left = alignment.center_left
    Center = alignment.center
    Right = alignment.center_right


class Location(Enum):
    """A relative location for placing elements or windows, that can go into any
    location
    """

    TopLeft = alignment.top_left
    TopCenter = alignment.top_center
    TopRight = alignment.top_right
    CenterLeft = alignment.center_left
    Center = alignment.center
    CenterRight = alignment.center_right
    BottomLeft = alignment.bottom_left
    BottomCenter = alignment.bottom_center
    BottomRight = alignment.bottom_right


SupportedFletLayout = Union[Row, Column, Container, Stack]


class PageNotOpenError(RuntimeError):
    """Raised when a method is called that requires the dialog to be open but dialog
    was not yet open"""


class LayoutError(ValueError):
    """Raised when an invalid layout is made. Debug, and do not catch these."""



from typing import Any, Dict, List, Optional, Tuple, Union

from typing_extensions import Literal
from RPA.core.types import is_list_like  # type: ignore

from RPA.Assistant.types import Element, Location, Options


def to_options(
    opts: Options, default: Optional[str] = None
) -> Tuple[List[str], Optional[str]]:
    """Convert keyword argument for multiple options into
    a list of strings.

    Also handles default option validation.
    """
    if isinstance(opts, str):
        opts = [opt.strip() for opt in opts.split(",")]
    elif is_list_like(opts):
        opts = [str(opt) for opt in opts]
    else:
        raise ValueError(f"Unsupported options type: {opts}")

    if not opts:
        return [], None

    if default is None:
        default = opts[0]

    if default not in opts:
        raise ValueError(f"Default '{default}' is not in available options")

    return opts, default


def optional_str(val: Any) -> Optional[str]:
    """Convert value to string, but keep NoneType"""
    return str(val) if val is not None else val


def optional_int(val: Any) -> Optional[int]:
    """Convert value to int, but keep NoneType"""
    return int(val) if val is not None else val


def int_or_auto(val: Any) -> Union[int, str]:
    """Convert value to int or 'AUTO' literal"""
    if isinstance(val, int):
        return val

    try:
        return int(val)
    except ValueError:
        pass

    height = str(val).strip().upper()
    if height == "AUTO":
        return height

    raise ValueError("Value not integer or AUTO")


def is_input(element: Element) -> bool:
    """Check if an element is an input"""
    return element["type"].startswith("input-")


def is_submit(element: Element) -> bool:
    """Check if an element is a submit button."""
    return element["type"] == "submit"


def location_to_absolute(
    location: Union[Location, Tuple[int, int], None],
    parent_width: float,
    parent_height: float,
    element_width: Optional[float],
    element_height: Optional[float],
) -> Dict[Literal["left", "top", "bottom", "right"], float]:
    """Calculates and returns absolute version of elements relative position as left or
    right and bottom or top keys in dictionary.
    """
    if isinstance(location, tuple):
        return {"left": location[0], "top": location[1]}

    if location in [
        Location.TopCenter,
        Location.BottomCenter,
        Location.CenterLeft,
        Location.CenterRight,
        Location.Center,
    ]:
        if element_height is None or element_width is None:
            raise ValueError(
                "Cannot determine centered position without static width and height"
            )
        half_height = (parent_height / 2) - (element_height / 2)
        half_width = (parent_width / 2) - (element_width / 2)
    else:
        half_height, half_width = -1, -1

    coordinates: Dict[
        Location, Dict[Literal["left", "top", "bottom", "right"], float]
    ] = {
        Location.TopLeft: {"left": 0, "top": 0},
        Location.TopCenter: {"left": half_width, "top": 0},
        Location.TopRight: {"right": 0, "top": 0},
        Location.CenterLeft: {"left": 0, "top": half_height},
        Location.Center: {"left": half_width, "top": half_height},
        Location.CenterRight: {"right": 0, "top": half_width},
        Location.BottomLeft: {"left": 0, "bottom": 0},
        Location.BottomCenter: {"left": half_width, "bottom": 0},
        Location.BottomRight: {"right": 0, "bottom": 0},
    }

    if isinstance(location, Location):
        return coordinates[location]
    else:
        raise ValueError(f"Invalid location {location}")



from dataclasses import dataclass
import time
from logging import getLogger
from timeit import default_timer as timer
from typing import Callable, Dict, List, Optional, Set, Tuple, Union

import flet
from flet import Container, Page, ScrollMode
from flet_core import Checkbox, Column, Control, AppBar, Row, Stack
from flet_core.control_event import ControlEvent
from typing_extensions import Literal

from RPA.Assistant.background_flet import BackgroundFlet
from RPA.Assistant.types import (
    LayoutError,
    PageNotOpenError,
    Result,
    SupportedFletLayout,
    WindowLocation,
)


def resolve_absolute_position(
    location: Union[WindowLocation, Tuple],
) -> Tuple[int, int]:

    if location is WindowLocation.TopLeft:
        return (0, 0)
    elif isinstance(location, tuple):
        return location
    else:
        raise ValueError(f"Invalid location {location}")


class TimeoutException(RuntimeError):
    """Timeout while waiting for dialog to finish."""


@dataclass
class Elements:
    """Lists of visible and invisible control elements"""

    visible: List[Control]
    visible_by_name: Dict[str, Control]
    invisible: List[Control]
    app_bar: Optional[AppBar]
    used_names: Set[str]


class FletClient:
    """Class for wrapping flet operations"""

    def __init__(self) -> None:
        self.logger = getLogger(__name__)
        self.results: Result = {}
        self.page: Optional[Page] = None
        self.pending_operation: Optional[Callable] = None

        self._elements: Elements = Elements([], {}, [], None, set())
        self._to_disable: List[flet.Control] = []
        self._layout_stack: List[Union[SupportedFletLayout, AppBar]] = []

        self._background_flet = BackgroundFlet()

    def _create_flet_target_function(
        self,
        title: str,
        height: Union[int, Literal["AUTO"]],
        width: int,
        on_top: bool,
        location: Union[WindowLocation, Tuple[int, int], None],
    ) -> Callable[[Page], None]:
        def inner_execute(page: Page):
            page.title = title
            if height != "AUTO":
                page.window_height = height
            page.window_width = width
            page.window_always_on_top = on_top

            # TODO: do we even allow None as argument?
            # or some Location.AUTO which would let OS handle position?
            if location is not None:
                if location is WindowLocation.Center:
                    page.window_center()
                else:
                    coordinates = resolve_absolute_position(location=location)
                    page.window_left = coordinates[0]
                    page.window_top = coordinates[1]
            page.scroll = ScrollMode.AUTO
            page.on_disconnect = lambda _: self._background_flet.close_flet_view()
            self.page = page
            self.update_elements()

        return inner_execute

    def _show_flet(
        self,
        target: Callable[[Page], None],
        timeout: int,
    ):
        self._background_flet.start_flet_view(target)
        view_start_time = timer()
        try:
            while self._background_flet.poll() is None:
                if callable(self.pending_operation):
                    self.pending_operation()  # pylint: disable=not-callable
                    self.pending_operation = None
                if timer() - view_start_time >= timeout:
                    self._background_flet.close_flet_view()
                    raise TimeoutException(
                        "Reached timeout while waiting for Assistant Dialog"
                    )
                time.sleep(0.1)
        except TimeoutException:
            # pylint: disable=raise-missing-from
            raise TimeoutException("Reached timeout while waiting for Assistant Dialog")
        finally:
            # Control's can't be re-used on multiple pages so we remove the page and
            # clear elements after flet closes
            self.page = None
            self.clear_elements()
            self._to_disable.clear()

    def _checkbox_data_saver(self, name: str, handler: Optional[Callable] = None):
        """_make_on_change_data_saver special case for checkboxes, until
        https://github.com/flet-dev/flet/issues/1251 is fixed"""

        def handle_string_bool(bool_string: Literal["true", "false"]):
            if bool_string == "true":
                return True
            elif bool_string == "false":
                return False
            else:
                raise ValueError(f"Invalid checkbox value {bool_string}")

        # We don't want the if inside the change_listener so it doesn't have to run on
        # every on_change event
        if not handler:

            def change_listener(e: ControlEvent):
                self.results[name] = handle_string_bool(e.data)

        else:

            def change_listener(e: ControlEvent):
                handler(e)
                self.results[name] = handle_string_bool(e.data)

        return change_listener

    def _make_on_change_data_saver(
        self, name: str, handler: Optional[Callable] = None
    ) -> Callable[[ControlEvent], None]:
        """Add flet event handler to record the element's data whenever content changes
        if ``handler`` is provided also call that.
        """

        # We don't want the if inside the change_listener so it doesn't have to run on
        # every on_change event
        if not handler:

            def change_listener(e: ControlEvent):
                self.results[name] = e.data

        else:

            def change_listener(e: ControlEvent):
                handler(e)
                self.results[name] = e.data

        return change_listener

    def _add_child_to_layout(self, child: Control):
        current_layout = self._layout_stack[-1]
        if isinstance(current_layout, (Row, Stack, Column)):
            current_layout.controls.append(child)
        elif isinstance(current_layout, AppBar):
            current_layout.actions.append(child)
        elif isinstance(current_layout, Container):
            if current_layout.content is not None:
                raise LayoutError("Attempting to place two content in one Container")
            current_layout.content = child
        else:
            raise RuntimeError("Unsupported layout element")

    def add_element(
        self,
        element: flet.Control,
        name: Optional[str] = None,
        extra_handler: Optional[Callable] = None,
        validation_func: Optional[Callable] = None,
    ):
        if name in self._elements.used_names:
            raise ValueError(f"Name `{name}` already in use")

        # if added element is a Container we don't create our own margin container to
        # not override added containers properties.
        if isinstance(element, Container):
            new_element = element
        else:
            # make a container that adds margin around the element
            new_element = Container(margin=5, content=element)

        if self._layout_stack:
            self._add_child_to_layout(new_element)
        else:
            self._elements.visible.append(new_element)

        if name is not None:
            self._elements.visible_by_name[name] = element
            self._elements.used_names.add(name)

            if isinstance(element, Checkbox):
                on_change_data_saver = self._checkbox_data_saver(name, extra_handler)

            else:
                on_change_data_saver = self._make_on_change_data_saver(
                    name, extra_handler
                )

            def on_change(event):
                if validation_func:
                    validation_func(event)
                on_change_data_saver(event)

            element.on_change = on_change

    def add_invisible_element(self, element: flet.Control, name: Optional[str] = None):
        self._elements.invisible.append(element)
        if name is not None:
            element.on_change = self._make_on_change_data_saver(name)

    def display_flet_window(
        self,
        title: str,
        height: Union[int, Literal["AUTO"]],
        width: int,
        on_top: bool,
        location: Union[WindowLocation, Tuple[int, int], None],
        timeout: int,
    ):
        self._show_flet(
            self._create_flet_target_function(title, height, width, on_top, location),
            timeout,
        )

    def clear_elements(self):
        self._elements = Elements([], {}, [], None, set())
        if self.page:
            if self.page.controls:
                self.page.controls.clear()
            self.page.overlay.clear()
            self.page.update()

    def update_elements(self):
        """Updates the UI and shows new elements which have been added into the element
        lists
        """
        if not self.page:
            raise PageNotOpenError("No page open when update_elements was called")

        # invisible elements have to be added before visible ones, for file pickers
        # to work correctly
        for element in self._elements.invisible:
            self.page.overlay.append(element)
        for element in self._elements.visible:
            self.page.add(element)
        if self._elements.app_bar:
            self.page.appbar = self._elements.app_bar
        self.page.update()

    def flet_update(self):
        """Runs a plain update of the flet UI, updating existing elements"""
        if not self.page:
            raise PageNotOpenError("Flet update called when page is not open")
        self.page.update()

    def lock_elements(self):
        for element in self._to_disable:
            element.disabled = True

    def unlock_elements(self):
        for element in self._to_disable:
            element.disabled = False

    def add_to_disablelist(self, element: flet.Control):
        """Added elements will be disabled when code is running from buttons"""
        self._to_disable.append(element)

    def set_title(self, title: str):
        """Set flet dialog title when it is running."""
        if not self.page:
            raise PageNotOpenError("Set title called when page is not open")
        self.page.title = title

    def add_layout(self, layout: SupportedFletLayout):
        """Add a layout element as the currently open layout element. Following
        add_element calls will add elements inside ``layout``."""
        self.add_element(layout)
        self._layout_stack.append(layout)

    def close_layout(self):
        """Stop adding layout elements to the latest opened layout"""
        self._layout_stack.pop()

    def set_appbar(self, app_bar: AppBar):
        if self._elements.app_bar:
            raise LayoutError("Only one navigation may be defined at a time")
        self._elements.app_bar = app_bar
        self._layout_stack.append(app_bar)

    def get_layout_dimensions(self) -> Tuple[Optional[float], Optional[float]]:
        if len(self._layout_stack) == 0:
            raise LayoutError("No parent element to determine dimensions from")
        current_layout = self._layout_stack[-1]
        if isinstance(current_layout, AppBar):
            raise LayoutError("Cannot use absolute positions in appbar")
        if current_layout.width is None or current_layout.height is None:
            raise RuntimeError("Cannot determine dimensions of parent element")

        return (current_layout.width, current_layout.height)

    def set_error(self, element_name: str, error: str):
        """Set an error for an element. The element must be a TextField or other
        Control that has error_text attribute."""
        element = self._elements.visible_by_name[element_name]
        if not hasattr(element, "error_text"):
            raise RuntimeError(
                "Tried to set error for element that does not have error_text attribute"
            )
        element.error_text = error
        self.flet_update()



import logging
import types
from typing import Any, Callable, Dict, Optional, Union

from flet_core import ControlEvent
from robot.errors import RobotError
from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
from robot.output.logger import LOGGER


class CallbackRunner:
    """provides helper functionality for running Robot and Python callbacks in
    RPA.Assistant."""

    def __init__(self, client) -> None:
        self.logger = logging.getLogger(__name__)
        self.validation_errors: Dict[str, Optional[str]] = {}
        self._client = client

    def _python_callback(
        self, function: Callable[[Any], None], *args, **kwargs
    ) -> Callable[[], None]:
        """wrapper code that is used to add wrapping for user functions when binding
        them to be run by buttons
        """

        def func_wrapper() -> None:
            return self._run_python_callback(function, *args, **kwargs)

        return func_wrapper

    def python_validation(
        self, element_name: str, function: Callable[[Any], Optional[str]]
    ) -> Callable[[ControlEvent], None]:
        """wrapper code that is used to add wrapping for user functions when binding
        them to be run on validations buttons
        """

        def validate(e: ControlEvent) -> None:
            error = self._run_python_callback(function, e.data)
            casted_error = str(error) if error else None
            e.control.error_text = casted_error
            self._client.flet_update()
            self.validation_errors[element_name] = casted_error

        return validate

    def _run_python_callback(self, function: Callable, *args, **kwargs):
        try:
            return function(*args, **kwargs)

        # This can be anything since it comes from the user function, we don't
        # want to let the user function crash the UI
        except Exception as err:  # pylint: disable=broad-except
            self.logger.error(f"Error calling Python function {function.__name__}")
            self.logger.error(err)
        finally:
            self._client.unlock_elements()
            self._client.flet_update()
        return None

    def _robot_callback(self, kw_name: str, *args, **kwargs) -> Callable[[], None]:
        """wrapper code that is used to add wrapping for user functions when binding
        them to be run by buttons
        """

        def func_wrapper() -> None:
            return self._run_robot_callback(kw_name, *args, **kwargs)

        return func_wrapper

    def robot_validation(
        self, element_name: str, kw_name: str
    ) -> Callable[[ControlEvent], None]:
        """wrapper code that is used to add wrapping for user functions when binding
        them to be run on validation
        """

        def validate(e: ControlEvent) -> None:
            def _nothing(*args, **kwargs):  # pylint: disable=unused-argument
                pass

            logging_methods = (LOGGER.start_keyword, LOGGER.end_keyword)
            try:
                # Disable logging by monkey patching the robot logger
                LOGGER.start_keyword = types.MethodType(_nothing, LOGGER)
                LOGGER.end_keyword = types.MethodType(_nothing, LOGGER)

                error = self._run_robot_callback(kw_name, e.control.value)
                e.control.error_text = error
                self._client.flet_update()
                self.validation_errors[element_name] = error
            finally:
                LOGGER.start_keyword, LOGGER.end_keyword = logging_methods

        return validate

    def _run_robot_callback(self, kw_name: str, *args, **kwargs):

        try:
            return BuiltIn().run_keyword(kw_name, *args, **kwargs)
        except RobotNotRunningError:
            self.logger.error(
                f"Robot Framework not running so cannot call keyword {kw_name}"
            )
        except RobotError as e:
            self.logger.error(f"Error calling robot keyword {kw_name}")
            self.logger.error(e)
        # This can be anything since it comes from the user function, we don't
        # want to let the user function crash the UI
        except Exception as err:  # pylint: disable=broad-except
            self.logger.error(f"Unexpected error running robot keyword {kw_name}")
            self.logger.error(err)
        finally:
            self._client.unlock_elements()
            self._client.flet_update()
        return None

    def queue_fn_or_kw(self, function: Union[Callable, str], *args, **kwargs):
        """Check if function is a Python function or a Robot Keyword, and schedule it
        for execution appropriately.
        """
        if self._client.pending_operation:
            self.logger.error("Can't have more than one pending operation.")
            return
        self._client.lock_elements()
        self._client.flet_update()

        if isinstance(function, Callable):
            self._client.pending_operation = self._python_callback(
                function, *args, **kwargs
            )
        else:
            self._client.pending_operation = self._robot_callback(
                function, *args, **kwargs
            )
